#/usr/bin/python

# Create table doc_state_time that contains the document index ID and the
# time in days in each document state for all the documents in the date
# range. The new table uses the InnoDB engine with transaction support
# so inserting, changing, or deleting rows needs a commit statement.
#
# The state_history table contains entries when documents enter new states,
# the idea is to loop through the entries for a given document accumulating
# the time by subtraction the current date from the previous one and
# assigning it to the previous state.
#
# The programs accepts the date range and the name of the file containing the
# database connection information on the command line. See functions
# usage() and cmdlineopts() in module sla_util for details. Database
# connection information is in get_db_creds().

# This is the default name of the file with the database connection info.
# Keep near the top for easy edit by site admin.

def set_default_db_cred():
    database_credentials = "db_creds"
    return database_credentials

import getopt
import sys
import datetime
import mysql.connector
import sla_util

################################################################################

def get_states(cnx):
    """Read the state names and numbers to store in dictionaries."""
    # At times we need the state_id numbers as keys and at other times
    # we do the lookup by name so create two dictionaries with one being
    # the transpose of the other.
    #
    # states has the id numbers as keys and the state names as values.
    # states_t has it the other way.
    cursor = cnx.cursor()
    states = {}
    states_t = {}
    try:
        cursor.execute("select * from states")
    except mysql.connector.Error as err:
        print "Select * from states failed: {}".format(err)
        exit(1)
    for x in cursor:
        states[x[0]] = x[1]
        states_t[x[1]] = x[0]
    cursor.close()
    return (states, states_t)

def create_doctime_table(cnx):
    """Create or recreate the doc_state_time table."""
    cursor = cnx.cursor()
    try:
        cursor.execute("drop table if exists doc_state_time")
    except mysql.connector.Error as err:
        print "Drop doc_state_time failed: {}".format(err)
        exit(1)
    create_stmt = (
        "create table doc_state_time ("
        "id int primary key,"
        "auth int,"
        "auth48 int,"
        "auth48_done int,"
        "dnp int,"
        "edit int,"
        "iana int,"
        "iesg int,"
        "isr int,"
        "isr_auth int,"
        "missref int,"
        "pending int,"
        "pub int,"
        "ref int,"
        "repl int,"
        "rfc_editor int,"
        "timeout int,"
        "withdrawn int) ENGINE=InnoDB")
    try:
        cursor.execute(create_stmt)
    except mysql.connector.Error as err:
        print "Create doc_state_time failed: {}".format(err)
        exit(1)
    cursor.close()
    return

def load_index_data(begin_date, end_date, cnx):
    """Fetch data from the index table for documents in date range. Return
a dictionay keyed on the document index value."""
    cursor = cnx.cursor()
    # Fetch information about documents in date range from the index table.
    #
    # Note: cursor.execute() knows how to convert python data types to
    # MySQL types better than .format() string method, especially for
    # dates.
    query = ("select internal_key, date_received, `doc-id`, `char-count`, "
             "`page-count`, `pub-date` from `index` where "
             "`pub-date` >= %s and `pub-date` <= %s")
    try:
        cursor.execute(query, (begin_date, end_date))
    except mysql.connector.Error as err:
        print "Select failed (q_index): {}".format(err)
        exit(1)
    # Document ID number is the key for idx dictionary.
    idx = {}
    for doc in cursor:
        idx[doc[0]] = doc[1:]
    cursor.close()
    return idx

def collapse_same_dates(state_hist):
    """ Take results from state_history query and put in date order.
    When there are entries on the same day, just use the last one
    from the unsorted query with that date."""

    # Yes, relying on result order of an unsorted query is an
    # exceptionally BAD idea. No choice in this case.

    # query results from "select state_id, in_date from state_history "
    #    "where internal_dockey = %s")

    results = []
    last_date = None
    multi_done = False
    # Sort state_hist by second element to loop in order of in_date.
    for line in sorted(state_hist, key=lambda z: z[1]):
        if line[1] == last_date:
            # Multiple entries on same day.
            # Check for having done this one already.
            if not multi_done:
                # Search backward to find the last one.
                for i in reversed(xrange(len(state_hist))):
                    if state_hist[i][1] == last_date:
                        break
                # Already found one for this day since we know it is
                # multiple. Replace the that with the last entry for this
                # day.
                results[-1] = [state_hist[i][0], last_date]
                multi_done = True
        else:
            multi_done = False
            last_date = line[1]
            results.append(line)
    return results

def build_insert(doc, line):
    """Takes the documnet ID and the state times to build the insert line."""
    # Convert the integer times for each state into a list of strings for
    # joining into one string.
    str_times = ["{0}".format(x) for x in line[5:]]
    insert_q = "insert into doc_state_time values ({0}, ".format(doc) + \
        ", ".join(str_times) + ")"
    return insert_q

def fetch_doc_history(doc, index_data, cursor, states):
    """Get state history for doc and compute days in each state."""
    # Need to process in order of in_date, but when there are
    # multiple entries on same day, then use default order from
    # select statement.
    statetime = {}
    prev_state = None
    prev_date = None
    sh_query = ("select state_id, in_date from state_history "
                "where internal_dockey = {}".format(doc))
    try:
        cursor.execute(sh_query)
    except mysql.connector.Error as err:
        print "State history query failed: {}".format(err)
        print "Statement: {}".format(cursor.statement)
        exit(1)
    results = collapse_same_dates(cursor.fetchall())
    for line in results:
        # Skip results with invalid state
        if line[0] in states.keys():
            if prev_date == None:
                # First time through
                prev_state = line[0]
                prev_date = line[1]
                continue
            if line[1] == prev_date:
                # Date same as last, so update state and date but don't do
                # anything else.
                prev_state = line[0]
                prev_date = line[1]
                continue
            # Passed sanity checks. Add time diff and update state.
            if not prev_state in statetime:
                statetime[prev_state] = 0
            statetime[prev_state] += (line[1] - prev_date).days
            prev_state = line[0]
            prev_date = line[1]
        else:
            print "Invalid state ({0}) found for doc_id {1}".format(line[0], 
                                                                  doc)
    # Check for time gap from last state history entry to pub date. Add
    # time if gap is there and positive.
    if index_data[4] and prev_date:
        if index_data[4] > prev_date:
            if not prev_state in statetime:
                statetime[prev_state] = 0
            statetime[prev_state] += (index_data[4] - prev_date).days
    return statetime

def process_doc(doc, index_data, cnx, states, flipped_states):
    """Write times in each state for this doc to database."""
    cursor = cnx.cursor()
    statetime = fetch_doc_history(doc, index_data, cursor, states)
    # Start line with doc ID, date received, pub date, char count, page count
    line = [index_data[1]]
    for x in [index_data[0], index_data[4], index_data[2], index_data[3]]:
        if x == None:
            line.append('NA')
        else:
            line.append(x)
    # Append days in each state to line.
    for k in sorted(flipped_states.keys()):
        state_id = flipped_states[k]
        if state_id in statetime:
            line.append(statetime[state_id])
        else:
            line.append(0)
    # Insert line into database.
    sql_stmt = build_insert(doc, line)
    try:
        cursor.execute(sql_stmt)
    except mysql.connector.Error as err:
        print "Doc time insert failed: {}".format(err)
        exit(1)
    cursor.close()
    return

################################################################################

def main():
    # Default begin and end dates are before earliest document for the start
    # and a future date for the end.
    database_credentials = set_default_db_cred()
    begin_date, end_date, new_db_creds = sla_util.cmdline_opts()
    if new_db_creds:
        database_credentials = new_db_creds
    cnx = sla_util.db_conn(database_credentials)
    states, states_t = get_states(cnx)
    idx = load_index_data(begin_date, end_date, cnx)
    # Set up doc_time table
    create_doctime_table(cnx)
    # For future version, generate all the rows ahead of time (probably to
    # file) and do one insert...values(...) statement. Until then use
    # a single transaction and commit at the end.
    try:
        cnx.start_transaction()
    except mysql.connector.Error as err:
        print "Failed to start transaction: {}".format(err)
        exit(1)
    # Loop through the documents.
    for doc_id in sorted(idx.keys()):
        process_doc(doc_id, idx[doc_id], cnx, states, states_t)
    try:
        cnx.commit()
    except mysql.connector.Error as err:
        print "Failed to commit transaction: {}".format(err)
        exit(1)
    cnx.close()
    exit(0)

if __name__ == "__main__":
    main()


