#!/usr/bin/python

# Note: SLA uses business days but ignores holidays so just
# scale thresholds to get calendar days.
# 30 business days = 42 calendar days
# 60 business days = 84 calendar days

#Option 1: Three tiers.
#
#All numbers below are by quarter.
#
#(i) When there is a normal amount of input, the SLA is 67% of
#documents published within the period have an RFC Editor controlled
#time that adds up to 30 business days or fewer.
#Where normal is defined as less than 1950 WF.
#
#(ii) When there is a moderate burst in the amount of input, then the
#SLA shifts to 50% of documents published within the period have an RFC
#Editor controlled time that adds up to 60 business days or fewer.
#Where a moderate burst is defined as 1950 - 3072 (inclusive) WF.
#
#(iii) When there is a large burst in the amount of input, then the SLA
#must be discussed and renegotiated.
#Where large burst is defined as greater than 3072 WF.
#
#- -----
#Programming logic can be whatever works for you.

#Some of the squishy things that are giving us difficulty:
#1. The IETF Chair wants the percentage in option (i) to be a variable
#that covers 90% of the typical scenarios.  I don't know if "67% of
#documents published" actually covers 90% of the use cases, of if that
#should be "80% of documents published" or "50% of documents published".

option_1_percentile = 0.67

# Change this value as needed for Option 1.

import getopt
import sys
import datetime
import mysql.connector
import math

# This is the default name of the file with the database connection info.
# Keep near the top for easy edit by site admin.

def set_default_db_cred():
    database_credentials = "db_creds"
    return database_credentials

################################################################################

def usage():
    """Prints short help message."""
    help_str =  "{} --begin=startdate --end=enddate ".format(sys.argv[0]) + \
    "[--dbfile=db_credentials_file]"
    print help_str
    print "    Date format is 'yyyy-mm-dd'"
    print

def cmdline_opts():
    """Interpret command line options."""
    try:
        opts, args = getopt.getopt(sys.argv[1:], '', [
            'begin=', 'end=', 'dbfile='])
    except getopt.GetoptError as err:
        print str(err)
        usage()
        sys.exit(2)
    database_credentials = set_default_db_cred()
    begin_date = None
    end_date = None
    for o, a in opts:
        if o == '--begin':
            y, m, d = [int(x) for x in a.split('-')]
            begin_date = datetime.date(y, m, d)
        elif o == '--end':
            y, m, d = [int(x) for x in a.split('-')]
            end_date = datetime.date(y, m, d)
        elif o == '--dbfile':
            database_credentials = a
    if not (begin_date and end_date):
        print "Both begin date and end date are needed"
        sys.exit(2)

    return (begin_date, end_date, database_credentials)

def get_db_creds(fname):
    """Reads database user, password, database name, host and port from a
text file. Host and port default to 127.0.0.1 and 3306.
Format is keyword: value. For example,
user: testaccount
password: dummy_pw
database: testDB
host: 127.0.0.1
port: 3306"""

    db_user = ''
    db_pw = ''
    db_name = ''
    db_host = '127.0.0.1'
    db_port = 3306
    with open(fname, 'r') as f:
        for line in f:
            # Check for comment.
            if line[0] == '#':
                continue
            # Separate key and value at first : character.
            # Value can contain :, but key cannot.
            colon_idx = line.find(':')
            k = line[:colon_idx].strip()
            v = line[colon_idx+1:].strip()
            if k == "user":
                db_user = v
            elif k == "database":
                db_name = v
            elif k == "password":
                db_pw = v
            elif k == "host":
                db_host = v
            elif k == "port":
                db_port = v
    return (db_user, db_pw, db_name, db_host, db_port)

def db_conn(fname):
    """Establish database connection using credentials from file. Sets
autocommit to False."""
    user, pw, db, host, port = get_db_creds(fname)
    try:
        conn = mysql.connector.connect(user=user, password=pw,
                                       host=host, database=db,
                                       port=port, autocommit=False)
    except mysql.connector.Error as err:
          if err.errno == errorcode.ER_ACCESS_DENIED_ERROR:
              print("Something is wrong with your user name or password")
          elif err.errno == errorcode.ER_BAD_DB_ERROR:
              print("Database does not exists")
          else:
              print(err)
    return conn

def rpc_time(doc, cnx):
    """Compute RPC-controlled time for given document"""
    cursor = cnx.cursor()
    query = ("select edit + rfc_editor from doc_state_time "
             "where id = {}".format(doc))
    try:
        cursor.execute(query)
    except mysql.connector.Error as err:
        print "RPC-controlled time query failed: {}".format(err)
        exit(1)
    result = cursor.fetchone()[0]
    cursor.close()
    return result

def get_docs_published(begin_date, end_date, cnx):
    """Get internal ID numbers for document published in date range."""
    cursor = cnx.cursor()
    query = ("select internal_key from `index` where `pub-date` >= %s "
             "and `pub-date` <= %s")
    try:
             cursor.execute(query, (begin_date, end_date))
    except mysql.connector.Error as err:
        print "Get docs published query failed: {}".format(err)
        exit(1)
    result = [line[0] for line in cursor]
    cursor.close()
    return result

def pages_gone_to_edit(begin_date, end_date, cnx):
    """Compute total page count for document enting EDIT state during
time period."""
    cursor = cnx.cursor()
    query = ("select sum(`page-count`) from `index` "
             "join (select distinct internal_dockey from state_history "
             "where in_date >= %s and in_date <= %s and state_id = 4) sh "
             "on `index`.internal_key = sh.internal_dockey")
    try:
        cursor.execute(query, (begin_date, end_date))
    except mysql.connector.Error as err:
        print "Get pages_in_edit query failed: {}".format(err)
        exit(1)
    result = cursor.fetchone()[0]
    cursor.close()
    return result

def compute_stats(rpc, p):
    """input: dictionary with rpc time for each document, percentile as
fraction (eg. 0.67 for 67th percentile)
    output: statistics in dictionary"""
    count = float(len(rpc))
    sorted_times = sorted(rpc.itervalues())
    mean = math.fsum(sorted_times) / count
    sumsq = 0.0
    for x in sorted_times:
        diff = x - mean
        sumsq = diff * diff
    # Population variance and standard deviation
    varpop = sumsq / count
    stddevpop = math.sqrt(varpop)
    # Median
    n = int(count)           # Need integer for list index
    if n % 2:
        # Odd number, median is middle value
        median = float(sorted_times[n/2])
    else:
        # Even number, need 1/2 between two middle values
        median = float((sorted_times[n/2] + 
                        sorted_times[n/2 - 1])) / 2.0
    # Find the 67th percentile using NIST definition of percentile,
    # http://www.itl.nist.gov/div898/handbook/prc/section2/prc252.htm
    # Engineering Statistics Handbook
    temp = p * (count + 1)
    k = int(math.floor(temp))
    d = k - temp
    percentile = sorted_times[k] + \
        d * (float(sorted_times[k + 1]) - float(sorted_times[k]))
    return {'count': n, 'mean': mean, 'median': median, 'variance': varpop,
                'stddev': stddevpop, 'percentile': percentile}

def sla_option_1(rpc, pgte):
    stats = compute_stats(rpc, option_1_percentile)
    print "Pages Gone To Edit (work function): {}".format(pgte)
    print "Number of documents published: {}".format(stats['count'])
    print "Mean: {} calendar days".format(stats['mean'])
    print "Median: {} calendar days".format(stats['median'])
    print "\n"
    print "Option 1"
    print
    percent_threshold = int(100 * option_1_percentile)
    print ("{}th percentile: ".format(percent_threshold) +
           "{} calendar days".format(stats['percentile']))
    print "Population variance: {}".format(stats['variance'])
    print "Population std. deviation: {} calendar days".format(stats['stddev'])
    if pgte > 3072:
        print "PGTE too large({}), SLA undefined".format(pgte)
    else:
        if pgte < 1950:
            sla_met = stats['percentile'] <= 42
            print "SLA: {}% <= 30 business days".format(percent_threshold)
        elif pgte <= 3072:
            sla_met = stats['median'] <= 84
            print "SLA: 50% <= 60 business days"
        print "SLA met: {}".format(sla_met)
    print "\n"
    return

def sla_option_2(rpc, pgte):
    #Option 2: Two tiers
    #
    #All numbers below are by quarter.
    #
    #(i) When there is a normal amount of input, the SLA is
    #90% of documents published within the period have an RFC Editor
    #controlled time that adds up to 60 business days or less.
    #Where normal is defined as less than 3072 WF.
    #
    #(ii) When there is a burst in the amount of input, then the SLA
    #(over a year period) must be discussed and renegotiated
    #where a burst is defined as greater than or equal to 3072 WF.
    #
    stats = compute_stats(rpc, 0.90)
    print "Option 2"
    print
    print "90th percentile: {} calendar days".format(stats['percentile'])
    if pgte < 3072:
        sla_met = stats['percentile'] <= 84
        print "SLA met (90% <= 60 business days): {}".format(sla_met)
    else:
        print "PGTE too large({}), SLA undefined".format(pgte)
    print
    return

def main():
    begin_date, end_date, db_creds = cmdline_opts()
    cnx = db_conn(db_creds)
    # Build dictionary of internal document ID numbers and RPC-controlled time.
    rpc = {doc: rpc_time(doc, cnx) for doc in
           get_docs_published(begin_date, end_date, cnx)}
    # rpc now has RPC-controlled time (calendar days) for each document
    # pgte was called "work function" in original discussion of SLA.
    pgte = pages_gone_to_edit(begin_date, end_date, cnx)
    # Done with database.
    cnx.close()
    print "{} to {}".format(begin_date, end_date)
    sla_option_1(rpc, pgte)
    sla_option_2(rpc, pgte)        
    exit(0)

if __name__ == "__main__":
    main()
